use crate::api::exec_query::protocol::{EncodedOptionalData, OptionalData, RequiredData};
use crate::cartridge::config::StorageConfiguration;
use crate::cartridge::update_tracing;
use sbroad::backend::sql::ir::{PatternWithParams, TmpSpaceMap};
use sbroad::errors::{Action, Entity, SbroadError};
use sbroad::executor::bucket::Buckets;
use sbroad::executor::engine::Configuration;
use sbroad::executor::ir::QueryType;
use sbroad::executor::lru::{Cache, LRUCache, DEFAULT_CAPACITY};
use sbroad::ir::value::Value;
use sbroad::otm::child_span;
use sbroad::{debug, error, warn};
use sbroad_proc::otm_child_span;
use std::any::Any;
use std::cell::{Ref, RefCell};
use std::fmt::Display;
use tarantool::tlua::LuaFunction;
use tarantool::tuple::Tuple;

struct Statement {
    id: u32,
    pattern: String,
}

#[derive(Default)]
struct PreparedStmt(Option<Statement>);

impl PreparedStmt {
    /// Extract prepared statement from the cache.
    ///
    /// # Errors
    /// - Returns None instead of a regular statement (sentinel node in the cache).
    fn statement(&self) -> Result<&Statement, SbroadError> {
        self.0.as_ref().ok_or_else(|| {
            SbroadError::Invalid(
                Entity::Statement,
                Some("Statement is not prepared".to_string()),
            )
        })
    }

    fn id(&self) -> Result<u32, SbroadError> {
        Ok(self.statement()?.id)
    }

    fn pattern(&self) -> Result<&str, SbroadError> {
        Ok(&self.statement()?.pattern)
    }
}

impl std::fmt::Debug for PreparedStmt {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        if let Some(ref stmt) = self.0 {
            write!(f, "PreparedStmt {:?}", stmt.pattern)
        } else {
            write!(f, "PreparedStmt None")
        }
    }
}

#[allow(clippy::module_name_repetitions)]
pub struct StorageRuntime {
    metadata: RefCell<StorageConfiguration>,
    cache: RefCell<LRUCache<String, PreparedStmt>>,
}

impl Configuration for StorageRuntime {
    type Configuration = StorageConfiguration;

    fn cached_config(&self) -> Result<Ref<Self::Configuration>, SbroadError> {
        self.metadata.try_borrow().map_err(|e| {
            SbroadError::FailedTo(Action::Borrow, Some(Entity::Metadata), format!("{e:?}"))
        })
    }

    fn clear_config(&self) -> Result<(), SbroadError> {
        let mut metadata = self.metadata.try_borrow_mut().map_err(|e| {
            SbroadError::FailedTo(Action::Borrow, Some(Entity::Metadata), format!("{e:?}"))
        })?;
        *metadata = StorageConfiguration::default();
        Ok(())
    }

    fn is_config_empty(&self) -> Result<bool, SbroadError> {
        let metadata = self.metadata.try_borrow().map_err(|e| {
            SbroadError::FailedTo(Action::Borrow, Some(Entity::Metadata), format!("{e:?}"))
        })?;
        Ok(metadata.is_empty())
    }

    fn get_config(&self) -> Result<Option<Self::Configuration>, SbroadError> {
        if self.is_config_empty()? {
            let lua = tarantool::lua_state();

            let storage_cache_capacity: LuaFunction<_> =
                lua.eval("return get_storage_cache_capacity;").unwrap();
            let capacity: u64 = match storage_cache_capacity.call() {
                Ok(capacity) => capacity,
                Err(e) => {
                    error!(
                        Option::from("getting storage cache capacity"),
                        &format!("{e:?}"),
                    );
                    return Err(SbroadError::LuaError(format!("{e:?}")));
                }
            };
            let storage_capacity = usize::try_from(capacity)
                .map_err(|e| SbroadError::Invalid(Entity::Cache, Some(format!("{e:?}"))))?;

            let storage_cache_size_bytes: LuaFunction<_> =
                lua.eval("return get_storage_cache_size_bytes;").unwrap();
            let cache_size_bytes = match storage_cache_size_bytes.call::<u64>() {
                Ok(size_bytes) => size_bytes,
                Err(e) => {
                    error!(
                        Option::from("getting storage cache size bytes"),
                        &format!("{e:?}"),
                    );
                    return Err(SbroadError::LuaError(format!("{e:?}")));
                }
            };
            let storage_size_bytes = usize::try_from(cache_size_bytes)
                .map_err(|e| SbroadError::Invalid(Entity::Cache, Some(format!("{e}"))))?;

            let jaeger_agent_host: LuaFunction<_> =
                lua.eval("return get_jaeger_agent_host;").unwrap();
            let jaeger_host: String = match jaeger_agent_host.call() {
                Ok(res) => res,
                Err(e) => {
                    error!(Option::from("getting jaeger agent host"), &format!("{e:?}"),);
                    return Err(SbroadError::LuaError(format!("{e:?}")));
                }
            };

            let jaeger_agent_port: LuaFunction<_> =
                lua.eval("return get_jaeger_agent_port;").unwrap();
            let jaeger_port: u16 = match jaeger_agent_port.call() {
                Ok(res) => res,
                Err(e) => {
                    error!(Option::from("getting jaeger agent port"), &format!("{e:?}"),);
                    return Err(SbroadError::LuaError(format!("{e:?}")));
                }
            };

            let mut metadata = StorageConfiguration::new();
            metadata.storage_capacity = storage_capacity;
            metadata.storage_size_bytes = storage_size_bytes;
            metadata.jaeger_agent_host = jaeger_host;
            metadata.jaeger_agent_port = jaeger_port;
            update_tracing(&metadata.jaeger_agent_host, metadata.jaeger_agent_port)?;

            return Ok(Some(metadata));
        }
        Ok(None)
    }

    fn update_config(&self, metadata: Self::Configuration) -> Result<(), SbroadError> {
        let mut cached_metadata = self.metadata.try_borrow_mut().map_err(|e| {
            SbroadError::FailedTo(Action::Borrow, Some(Entity::Metadata), format!("{e:?}"))
        })?;
        let storage_size_bytes = metadata.storage_size_bytes;
        *cached_metadata = metadata;
        update_box_param("sql_cache_size", storage_size_bytes);
        Ok(())
    }
}

impl StorageRuntime {
    /// Build a new storage runtime.
    ///
    /// # Errors
    /// - Failed to initialize the LRU cache.
    pub fn new() -> Result<Self, SbroadError> {
        let cache: LRUCache<String, PreparedStmt> =
            LRUCache::new(DEFAULT_CAPACITY, Some(Box::new(unprepare)))?;
        let result = StorageRuntime {
            metadata: RefCell::new(StorageConfiguration::new()),
            cache: RefCell::new(cache),
        };

        Ok(result)
    }

    #[allow(unused_variables)]
    pub fn execute_plan(
        &self,
        required: &mut RequiredData,
        raw_optional: &mut Vec<u8>,
    ) -> Result<Box<dyn Any>, SbroadError> {
        if required.can_be_cached {
            return self.execute_cacheable_plan(required, raw_optional);
        }
        Self::execute_non_cacheable_plan(required, raw_optional)
    }

    #[allow(unused_variables)]
    pub fn execute_non_cacheable_plan(
        required: &mut RequiredData,
        raw_optional: &mut Vec<u8>,
    ) -> Result<Box<dyn Any>, SbroadError> {
        let plan_id = required.plan_id.clone();

        if required.can_be_cached {
            return Err(SbroadError::Invalid(
                Entity::Plan,
                Some("Expected a plan that can not be cached.".to_string()),
            ));
        }

        let (pattern_with_params, _tmp_spaces) = compile_encoded_optional(raw_optional)?;
        debug!(
            Option::from("execute"),
            &format!(
                "Failed to execute the statement: {}",
                pattern_with_params.pattern
            ),
        );
        let result = if required.query_type == QueryType::DML {
            write_unprepared(&pattern_with_params.pattern, &pattern_with_params.params)
        } else {
            read_unprepared(&pattern_with_params.pattern, &pattern_with_params.params)
        };

        result
    }

    #[allow(unused_variables)]
    pub fn execute_cacheable_plan(
        &self,
        required: &mut RequiredData,
        raw_optional: &mut Vec<u8>,
    ) -> Result<Box<dyn Any>, SbroadError> {
        let plan_id = required.plan_id.clone();

        if !required.can_be_cached {
            return Err(SbroadError::Invalid(
                Entity::Plan,
                Some("Expected a plan that can be cached.".to_string()),
            ));
        }

        // Look for the prepared statement in the cache.
        if let Some(stmt) = self
            .cache
            .try_borrow_mut()
            .map_err(|e| {
                SbroadError::FailedTo(Action::Borrow, Some(Entity::Cache), format!("{e}"))
            })?
            .get(&plan_id)?
        {
            let stmt_id = stmt.id()?;
            // The statement was found in the cache, so we can execute it.
            debug!(
                Option::from("execute plan"),
                &format!("Execute prepared statement: {stmt:?}"),
            );
            let result = match required.query_type {
                QueryType::DML => write_prepared(stmt_id, "", &required.parameters),
                QueryType::DQL => read_prepared(stmt_id, "", &required.parameters),
            };

            // If prepared statement is invalid for some reason, fallback to the long pass
            // and recompile the query.
            if result.is_ok() {
                return result;
            }
        }
        debug!(
            Option::from("execute plan"),
            &format!("Failed to find a plan (id {plan_id}) in the cache."),
        );

        let (pattern_with_params, _tmp_spaces) = compile_encoded_optional(raw_optional)?;
        let result = match prepare(&pattern_with_params.pattern) {
            Ok(stmt) => {
                let stmt_id = stmt.id()?;
                debug!(
                    Option::from("execute plan"),
                    &format!(
                        "Created prepared statement {} for the pattern {}",
                        stmt_id,
                        stmt.pattern()?
                    ),
                );
                self.cache
                    .try_borrow_mut()
                    .map_err(|e| {
                        SbroadError::FailedTo(
                            Action::Put,
                            None,
                            format!("prepared statement {stmt:?} into the cache: {e:?}"),
                        )
                    })?
                    .put(plan_id, stmt)?;
                // The statement was found in the cache, so we can execute it.
                debug!(
                    Option::from("execute plan"),
                    &format!("Execute prepared statement: {stmt_id}"),
                );
                if required.query_type == QueryType::DML {
                    write_prepared(
                        stmt_id,
                        &pattern_with_params.pattern,
                        &pattern_with_params.params,
                    )
                } else {
                    read_prepared(
                        stmt_id,
                        &pattern_with_params.pattern,
                        &pattern_with_params.params,
                    )
                }
            }
            Err(e) => {
                // Possibly the statement is correct, but doesn't fit into
                // Tarantool's prepared statements cache (`sql_cache_size`).
                // So we try to execute it bypassing the cache.
                warn!(
                    Option::from("execute"),
                    &format!(
                        "Failed to prepare the statement: {}, error: {e}",
                        pattern_with_params.pattern
                    ),
                );
                if required.query_type == QueryType::DML {
                    write_unprepared(&pattern_with_params.pattern, &pattern_with_params.params)
                } else {
                    read_unprepared(&pattern_with_params.pattern, &pattern_with_params.params)
                }
            }
        };

        result
    }
}

fn compile_encoded_optional(
    raw_optional: &mut Vec<u8>,
) -> Result<(PatternWithParams, TmpSpaceMap), SbroadError> {
    // Use all buckets as we don't want to filter any data from the execution plan
    // (this work has already been done on the coordinator).
    let buckets = Buckets::All;

    // Find a statement in the Tarantool's cache or prepare it
    // (i.e. compile and put into the cache).
    let data = std::mem::take(raw_optional);
    let mut optional = OptionalData::try_from(EncodedOptionalData::from(data))?;
    optional.exec_plan.get_mut_ir_plan().restore_constants()?;
    let nodes = optional.ordered.to_syntax_data()?;
    optional.exec_plan.to_sql(
        &nodes,
        &buckets,
        &uuid::Uuid::new_v4().as_simple().to_string(),
    )
}

#[otm_child_span("tarantool.statement.prepare")]
fn prepare(pattern: &str) -> Result<PreparedStmt, SbroadError> {
    let lua = tarantool::lua_state();

    let prepare_stmt: LuaFunction<_> = lua
        .get("prepare")
        .ok_or_else(|| SbroadError::LuaError("Lua function `prepare` not found".into()))?;

    match prepare_stmt.call_with_args::<u32, _>(pattern) {
        Ok(stmt_id) => {
            let stmt = Statement {
                id: stmt_id,
                pattern: pattern.to_string(),
            };
            Ok(PreparedStmt(Some(stmt)))
        }
        Err(e) => {
            error!(Option::from("prepare"), &format!("{e:?}"));
            Err(SbroadError::LuaError(format!("{e:?}")))
        }
    }
}

#[otm_child_span("tarantool.statement.unprepare")]
fn unprepare(stmt: &mut PreparedStmt) -> Result<(), SbroadError> {
    let lua = tarantool::lua_state();

    let unprepare_stmt: LuaFunction<_> = lua
        .get("unprepare")
        .ok_or_else(|| SbroadError::LuaError("Lua function `unprepare` not found".into()))?;

    match unprepare_stmt.call_with_args::<(), _>(stmt.id()?) {
        Ok(_) => Ok(()),
        Err(e) => {
            error!(Option::from("unprepare"), &format!("{e:?}"));
            Err(SbroadError::LuaError(format!("{e:?}")))
        }
    }
}

#[otm_child_span("tarantool.statement.prepared.read")]
fn read_prepared(stmt_id: u32, stmt: &str, params: &[Value]) -> Result<Box<dyn Any>, SbroadError> {
    let lua = tarantool::lua_state();

    let exec_sql: LuaFunction<_> = lua
        .get("read")
        .ok_or_else(|| SbroadError::LuaError("Lua function `read` not found".into()))?;

    match exec_sql.call_with_args::<Tuple, _>((stmt_id, stmt, params)) {
        Ok(v) => Ok(Box::new(v) as Box<dyn Any>),
        Err(e) => {
            error!(Option::from("read_prepared"), &format!("{e:?}"));
            Err(SbroadError::LuaError(format!("{e:?}")))
        }
    }
}

#[otm_child_span("tarantool.statement.unprepared.read")]
fn read_unprepared(stmt: &str, params: &[Value]) -> Result<Box<dyn Any>, SbroadError> {
    let lua = tarantool::lua_state();

    let exec_sql: LuaFunction<_> = lua
        .get("read")
        .ok_or_else(|| SbroadError::LuaError("Lua function `read` not found".into()))?;

    match exec_sql.call_with_args::<Tuple, _>((0, stmt, params)) {
        Ok(v) => Ok(Box::new(v) as Box<dyn Any>),
        Err(e) => {
            error!(Option::from("read_unprepared"), &format!("{e:?}"));
            Err(SbroadError::LuaError(format!("{e:?}")))
        }
    }
}

#[otm_child_span("tarantool.statement.prepared.write")]
fn write_prepared(stmt_id: u32, stmt: &str, params: &[Value]) -> Result<Box<dyn Any>, SbroadError> {
    let lua = tarantool::lua_state();

    let exec_sql: LuaFunction<_> = lua
        .get("write")
        .ok_or_else(|| SbroadError::LuaError("Lua function `write` not found".into()))?;

    match exec_sql.call_with_args::<Tuple, _>((stmt_id, stmt, params)) {
        Ok(v) => Ok(Box::new(v) as Box<dyn Any>),
        Err(e) => {
            error!(Option::from("write_prepared"), &format!("{e:?}"));
            Err(SbroadError::LuaError(format!("{e:?}")))
        }
    }
}

#[otm_child_span("tarantool.statement.unprepared.write")]
fn write_unprepared(stmt: &str, params: &[Value]) -> Result<Box<dyn Any>, SbroadError> {
    let lua = tarantool::lua_state();

    let exec_sql: LuaFunction<_> = lua
        .get("write")
        .ok_or_else(|| SbroadError::LuaError("Lua function `write` not found".into()))?;

    match exec_sql.call_with_args::<Tuple, _>((0, stmt, params)) {
        Ok(v) => Ok(Box::new(v) as Box<dyn Any>),
        Err(e) => {
            error!(Option::from("write_unprepared"), &format!("{e:?}"));
            Err(SbroadError::LuaError(format!("{e:?}")))
        }
    }
}

fn update_box_param<T>(param: &str, val: T)
where
    T: Display,
{
    let lua = tarantool::lua_state();
    match lua.exec(&format!("box.cfg{{{param} = {val}}}")) {
        Ok(_) => debug!(
            Option::from("update_box_param"),
            &format!("box.cfg param {param} was updated to {val}")
        ),
        Err(e) => warn!(
            Option::from("update_box_param"),
            &format!("box.cfg update error: {e}")
        ),
    }
}
